# -*- coding: utf-8 -*-

"""
This module provides a class :class:`Call` that allows you to use Oracle PL/SQL
procedures/functions as CherryPy__ response handlers. A :class:`Call` objects
wraps a :class:`ll.orasql.Procedure` or :class:`ll.orasql.Function` object from
the :mod:`ll.orasql` module.

__ http://www.cherrypy.org/

For example, you might have the following PL/SQL function::

	create or replace function helloworld
	(
		who varchar2
	)
	return varchar2
	as
	begin
		return '<html><head><h>Hello ' || who || '</h></head><body><h1>Hello, ' || who || '!</h1></body></html>';
	end;

Using this function as a CherryPy response handler can be done like this::

	import cherrypy

	from ll import orasql, nightshade


	proc = nightshade.Call(orasql.Function("helloworld"), connectstring="user/pwd")

	class HelloWorld(object):
		@cherrypy.expose
		def default(self, who="World"):
			cherrypy.response.headers["Content-Type"] = "text/html"
			return proc(who=who)

	cherrypy.quickstart(HelloWorld())
"""

import time, datetime, threading

import cherrypy

from ll import orasql


__docformat__ = "reStructuredText"

weekdayname = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
monthname = [None, "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]


class UTC(datetime.tzinfo):
	"""
	Timezone object for UTC
	"""
	def utcoffset(self, dt):
		return datetime.timedelta(0)

	def dst(self, dt):
		return datetime.timedelta(0)

	def tzname(self, dt):
		return "UTC"

utc = UTC()


def getnow():
	"""
	Get the current date and time as a :class:`datetime.datetime` object in UTC
	with timezone info.
	"""
	return datetime.datetime.utcnow().replace(tzinfo=utc)


def httpdate(dt):
	"""
	Return a string suitable for a "Last-Modified" and "Expires" header.

	:var:`dt` is a :class:`datetime.datetime` object. If ``:var:`dt`.tzinfo`` is
	:const:`None` :var:`dt` is assumed to be in the local timezone (using the
	current UTC offset which might be different from the one used by :var:`dt`).
	"""
	if dt.tzinfo is None:
		dt += datetime.timedelta(seconds=[time.timezone, time.altzone][time.daylight])
	else:
		dt -= dt.tzinfo.utcoffset(dt)
	return "{0}, {1:02d} {2:3s} {3:4d} {4:02d}:{5:02d}:{6:02d} GMT".format(weekdayname[dt.weekday()], dt.day, monthname[dt.month], dt.year, dt.hour, dt.minute, dt.second)


class Connect(object):
	"""
	:class:`Connect` objects can be used as decorators that wraps a function
	that needs a database connection.

	If calling the wrapped function results in a database exception that has
	been caused by a lost connection to the database or similar problems,
	the function is retried with a new database connection.
	"""
	_badoracleexceptions = set((
		28,    # your session has been killed
		1012,  # not logged on
		1014,  # Oracle shutdown in progress
		1033,  # Oracle startup or shutdown in progress
		1034,  # Oracle not available
		1035,  # Oracle only available to users with RESTRICTED SESSION privilege
		1089,  # immediate shutdown in progress - no operations are permitted
		1090,  # Shutdown in progress - connection is not permitted
		1092,  # ORACLE instance terminated. Disconnection forced
		3106,  # fatal two-task communication protocol error
		3113,  # end-of-file on communication channel
		3114,  # not connected to ORACLE
		3135,  # connection lost contact
		12154, # TNS:could not resolve the connect identifier specified
		12540, # TNS:internal limit restriction exceeded
		12541, # TNS:no listener
		12543, # TNS:destination host unreachable
	))

	def __init__(self, connectstring=None, pool=None, retry=3, **kwargs):
		"""
		Create a new parameterized :class:`Connect` decorator. Either
		:var:`connectstring` or :var:`pool` (a database pool object) must be
		specified. :var:`retry` specifies how often to retry calling the wrapped
		function after a database exception. :var:`kwargs` will be passed on to
		the :func:`connect` call.
		"""
		if (connectstring is not None) == (pool is not None):
			raise TypeError("either connectstring or pool must be specified")
		self.pool = pool
		self._connection = None
		self.connectstring = connectstring
		self.retry = retry
		self.kwargs = kwargs

	def _isbadoracleexception(self, exc):
		if exc.args:
			code = getattr(exc[0], "code", 0)
			if code in self._badoracleexceptions:
				return True
		return False

	def _getconnection(self):
		if self.pool is not None:
			return self.pool.acquire()
		elif self._connection is None:
			self._connection = orasql.connect(self.connectstring, threaded=True, **self.kwargs)
		return self._connection

	def _dropconnection(self, connection):
		if self.pool is not None:
			self.pool.drop(connection)
		else:
			self._connection = None

	def cursor(self, **kwargs):
		connection = self._getconnection()
		return connection.cursor(**kwargs)

	def commit(self):
		self._getconnection().commit()

	def rollback(self):
		self._getconnection().rollback()

	def close(self):
		connection = self._getconnection()
		connection.close()
		self._dropconnection(connection)

	def cancel(self):
		self._getconnection().cancel()

	def __call__(self, func):
		def wrapper(*args, **kwargs):
			for i in xrange(self.retry):
				connection = self._getconnection()
				try:
					# This only works if func is using the same connection
					return func(*args, **kwargs)
				except orasql.DatabaseError, exc:
					if i<self.retry-1 and self._isbadoracleexception(exc):
						# Drop bad connection and retry
						self._dropconnection(connection)
					else:
						raise
		wrapper.__name__ = func.__name__
		wrapper.__doc__ = func.__doc__
		wrapper.__dict__.update(func.__dict__)
		return wrapper


class Call(object):
	"""
	Wrap an Oracle procedure or function in a CherryPy handler.

	A :class:`Call` object wraps a procedure or function object from
	:mod:`ll.orasql` and makes it callable just like a CherryPy handler.
	"""
	def __init__(self, callable, connection):
		"""
		Create a :class:`Call` object wrapping the function or procedure
		:var:`callable`.
		"""
		self.callable = callable
		# Calculate parameter mapping now, so we don't get concurrency problems later
		self.connection = connection
		callable._calcargs(connection.cursor())

	def __call__(self, *args, **kwargs):
		"""
		Call the procedure/function with the arguments :var:`args` and
		:var:`kwargs` mapping Python function arguments to
		Oracle procedure/function arguments. On return from the procedure the
		:var:`c_out` parameter is mapped to the CherryPy response body, and the
		parameters :var:`p_expires` (the number of days from now),
		:var:`p_lastmodified` (a date in UTC), :var:`p_mimetype`: (a string),
		:var:`p_encoding` (a string), :var:`p_etag` (a string) and
		:var:`p_cachecontrol` (a string) are mapped to the appropriate CherryPy
		response headers. If :var:`p_etag` is not specified a value is calculated.

		If the procedure/function raised a PL/SQL exception with a code between
		20200 and 20599, 20000 will be substracted from this value and the
		resulting value will be used as the HTTP response code, i.e. 20404 will
		give a "Not Found" response.
		"""

		@self.connection
		def call(*args, **kwargs):
			cursor = self.connection.cursor()
			try:
				if isinstance(self.callable, orasql.Procedure):
					result = (None, self.callable(cursor, *args, **kwargs))
				else:
					result = self.callable(cursor, *args, **kwargs)
				cursor.connection.commit()
				return result
			except orasql.DatabaseError, exc:
				if exc.args:
					code = getattr(exc[0], "code", 0)
					if 20200 <= code <= 20599:
						raise cherrypy.HTTPError(code-20000)
					else:
						raise

		now = getnow()
		(body, result) = call(*args, **kwargs)

		# Set HTTP headers from parameters
		expires = result.get("p_expires", None)
		if expires is not None:
			cherrypy.response.headers["Expires"] = httpdate(now + datetime.timedelta(days=expires))
		lastmodified = result.get("p_lastmodified", None)
		if lastmodified is not None:
			cherrypy.response.headers["Last-Modified"] = httpdate(lastmodified)
		encoding = None
		if isinstance(result, unicode):
			encoding = "utf-8"
			result = result.encoding(encoding)
		mimetype = result.get("p_mimetype", None)
		if mimetype is not None:
			if encoding is None:
				encoding = result.get("p_encoding", None)
			if encoding is not None:
				cherrypy.response.headers["Content-Type"] = "{0}; charset={1}".format(mimetype, encoding)
			else:
				cherrypy.response.headers["Content-Type"] = mimetype
		hasetag = False
		etag = result.get("p_etag", None)
		if etag is not None:
			cherrypy.response.headers["ETag"] = etag
			hasetag = True
		cachecontrol = result.get("p_cachecontrol", None)
		if cachecontrol is not None:
			cherrypy.response.headers["Cache-Control"] = cachecontrol

		# Get status code
		status = result.get("p_status", None)
		if status is not None:
			cherrypy.response.status = status

		# Get response body
		if "c_out" in result:
			body = result.c_out
			if hasattr(result, "read"):
				result = result.read()
			if not hasetag:
				cherrypy.response.headers["ETag"] = '"{0:x}"'.format(hash(body))

		if hasattr(body, "read"):
			body = body.read()
		return body



#! /usr/bin/env/python
# -*- coding: utf-8 -*-

## Copyright 1999-2010 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2010 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/__init__.py for the license


from ll import url
from ll.xist import xsc, css, parse
from ll.xist.ns import html, specials


def test_css():
	with xsc.build():
		with html.div(id=1) as e:
			with html.ul(id=2):
				+html.li(u"foo")
				+html.li()

	assert list(e.walknodes(css.selector(u"div"))) == [e]
	assert list(e.walknodes(css.selector(u"li"))) == [e[0][0], e[0][1]]
	assert list(e.walknodes(css.selector(u"div#1"))) == [e]
	assert list(e.walknodes(css.selector(u"#2"))) == [e[0]]
	assert list(e.walknodes(css.selector(u":empty"))) == [e[0][1]]
	assert list(e.walknodes(css.selector(u"li:empty"))) == [e[0][1]]
	assert list(e.walknodes(css.selector(u"div :empty"))) == [e[0][1]]
	assert list(e.walknodes(css.selector(u"div>*:empty"))) == []
	assert list(e.walknodes(css.selector(u"div>:empty"))) == []
	assert list(e.walknodes(css.selector(u"*|li"))) == [e[0][0], e[0][1]]
	assert list(e.walknodes(css.selector(u"h|li", prefixes={u"h": html}))) == [e[0][0], e[0][1]]
	assert list(e.walknodes(css.selector(u"h|li", prefixes={u"h": specials}))) == []

	with xsc.build():
		with xsc.Frag() as e:
			+html.div(u"foo")
			+xsc.Text(u"filler")
			+html.p(u"foo")
			+xsc.Text(u"filler")
			+html.ul(html.li(u"foo"))

	assert list(e.walknodes(css.selector(u"div + p"))) == [e[2]]
	assert list(e.walknodes(css.selector(u"div + ul"))) == []
	assert list(e.walknodes(css.selector(u"ul + p"))) == []
	assert list(e.walknodes(css.selector(u"div ~ p"))) == [e[2]]
	assert list(e.walknodes(css.selector(u"div ~ ul"))) == [e[4]]
	assert list(e.walknodes(css.selector(u"p ~ div"))) == []
	assert list(e.walknodes(css.selector(u"div:first-child + p"))) == [e[2]]
	assert list(e.walknodes(css.selector(u"*:first-child + p"))) == [e[2]]

	with xsc.build():
		with xsc.Frag() as e:
			+html.span(html.b(u"hurz"), u"gurk", html.em(u"hinz"), html.em(u"kunz"))
			+html.em(u"hurz")
			+html.em(u"hinz")
			+xsc.Text(u"nix")
			+html.i(u"kunz")

	assert list(e.walknodes(css.selector(u"*:only-of-type"))) == [e[0], e[0][0], e[4]]
	assert list(e.walknodes(css.selector(u"*:nth-child(1)"))) == [e[0], e[0][0]]
	assert list(e.walknodes(css.selector(u"*:nth-child(2)"))) == [e[0][2], e[1]]
	assert list(e.walknodes(css.selector(u"*:nth-last-child(1)"))) == [e[0][3], e[4]]
	assert list(e.walknodes(css.selector(u"*:nth-last-child(2)"))) == [e[0][2], e[2]]
	assert list(e.walknodes(css.selector(u"*:nth-of-type(1)"))) == [e[0], e[0][0], e[0][2], e[1], e[4]]
	assert list(e.walknodes(css.selector(u"*:nth-of-type(2)"))) == [e[0][3], e[2]]
	assert list(e.walknodes(css.selector(u"*:nth-last-of-type(1)"))) == [e[0], e[0][0], e[0][3], e[2], e[4]]
	assert list(e.walknodes(css.selector(u"*:nth-last-of-type(2)"))) == [e[0][2], e[1]]

	e = xsc.Frag(html.span(html.b(u"hurz"), u"gurk"))
	assert list(e.walknodes(css.selector(u"*:only-child"))) == [e[0], e[0][0]]

	with xsc.build():
		with xsc.Frag() as e:
			+html.em(class_=u"gurk", lang=u"en")
			+html.em(class_=u"gurk hurz", lang=u"en-us")
			+html.em(class_=u"hurz", lang=u"de")

	assert list(e.walknodes(css.selector(u"em[class='gurk']"))) == [e[0]]
	assert list(e.walknodes(css.selector(u"em[class~='gurk']"))) == [e[0], e[1]]
	assert list(e.walknodes(css.selector(u"em[lang|='en']"))) == [e[0], e[1]]


def test_applystylesheets1():
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"p {color: red;}", type=u"text/css")
			with html.body():
				+html.p(u"gurk")

	css.applystylesheets(e)

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: red;"
	assert list(e.walknodes(html.style)) == []


def test_applystylesheets2():
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"p.dont {color: red;}", type=u"text/css")
			with html.body():
				+html.p(u"gurk")

	css.applystylesheets(e)

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u""
	assert list(e.walknodes(html.style)) == []


def test_applystylesheets3():
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"p.do {color: red;}", type=u"text/css")
			with html.body():
				+html.p(u"gurk", class_=u"do")

	css.applystylesheets(e)

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: red;"
	assert list(e.walknodes(html.style)) == []


def test_applystylesheets4():
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"#id42 {color: red;}", type=u"text/css")
			with html.body():
				+html.p(u"gurk", id=u"id42", style=u"color: blue;")

	css.applystylesheets(e)

	# style attribute wins
	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: blue;"
	assert list(e.walknodes(html.style)) == []


def test_applystylesheets5():
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"p#id42 {color: red;}", type=u"text/css")
			with html.body():
				+html.p(u"gurk", id=u"id42", style=u"color: blue;")

	css.applystylesheets(e)

	# stylesheet always wins (at least in CSS 2.1 and 3)
	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: blue;"
	assert list(e.walknodes(html.style)) == []


def test_applystylesheets_media():
	# Check that media="screen" picks up the media stylesheet
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"p {color: red;}", type=u"text/css", media=u"screen")
			with html.body():
				+html.p(u"gurk")

	css.applystylesheets(e, media=u"screen")

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: red;"

	# Check that media="screen" doesn't pick up the print stylesheet
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"p {color: red;}", type=u"text/css", media=u"screen")
			with html.body():
				+html.p(u"gurk")

	css.applystylesheets(e, media=u"print")

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u""

	# Check that @media rules are treated properly
	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"@media screen { p {color: red;}}", type=u"text/css")
			with html.body():
				+html.p(u"gurk")

	css.applystylesheets(e, media=u"screen")

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: red;"

	with xsc.build():
		with html.html() as e:
			with html.head():
				+html.style(u"@media screen { p {color: red;}}", type=u"text/css")
			with html.body():
				+html.p(u"gurk")

	css.applystylesheets(e, media=u"print")

	assert unicode(e.walknodes(html.p)[0].attrs.style) == u""


def test_applystylesheets_title():
	def makenode():
		with xsc.build():
			with html.html() as e:
				with html.head():
					+html.style(u"p {color: red;}", type=u"text/css")
					+html.style(u"p {color: blue;}", type=u"text/css", title=u"blue")
				with html.body():
					+html.p(u"gurk")
		return e

	# Check that title=None uses only the titleless stylesheets
	e = makenode()
	css.applystylesheets(e, title=None)
	assert unicode(e.walknodes(html.p)[0].attrs.style) == u"color: red;"

	# Check that title="blue" uses only the stylesheet with the specified title
	e = makenode()
	css.applystylesheets(e, title=u"blue")
	assert str(e.walknodes(html.p)[0].attrs.style) == u"color: blue;"


def test_parse():
	s = css.parsestring(b"@charset 'utf-8'; div{background-image: url(gurk.gif);}")
	urls = set(css.geturls(s))
	assert urls == set([url.URL("gurk.gif")])

	s = css.parsestring(b"@charset 'utf-8'; div{background-image: url(gurk.gif);}", base="root:")
	urls = set(css.geturls(s))
	assert urls == set([url.URL("root:gurk.gif")])


def test_comments():
	d = b'<html><head><style type="text/css">/*nix*/ p{/*nix*/ color: red;}</style></head><body><p>gurk</p></body></html>'
	node = parse.tree(d, parse.Expat(), parse.NS(html), parse.Node())
	css.applystylesheets(node)
	assert unicode(node.walknodes(html.p)[0].attrs.style) == u"color: red;"
